# import networkx as nx
#
# from imxInsights.graph.queries.sectionGeometryQuery import SectionGeometryGraphQuery
#
# # class GraphQueries:
# #     def __init__(self, g: nx.DiGraph):
# #         self.SectionGeometryGraphQuery: SectionGeometryGraphQuery = SectionGeometryGraphQuery(g)
